from jobio.job import submit
from jobio.cli.args import (
    add_execute_group,
    add_job_meta_group,
    add_bucket_group,
    add_s3_group,
    add_storage_group,
)


def add_job_cli(parser):
    job_commands = parser.add_subparsers(title="Commands")
    run_parser = job_commands.add_parser("run")
    add_job_meta_group(run_parser)
    add_execute_group(run_parser)
    add_storage_group(run_parser)
    add_s3_group(run_parser)
    add_bucket_group(run_parser)
    run_parser.set_defaults(func=submit)



import botocore
import boto3
import copy
import os


required_s3_fields = {"region_name": str}

required_s3_values = {"region_name": True}

valid_s3_resource_fields = ["region_name"]

required_bucket_fields = {
    "name": str,
    "input_prefix": str,
    "output_prefix": str,
}

required_bucket_values = {
    "name": True,
    "input_prefix": False,
    "output_prefix": False,
}


def upload_to_s3(s3_client, local_path, s3_path, bucket_name):
    if not os.path.exists(local_path):
        return False
    s3_client.upload_file(local_path, bucket_name, s3_path)
    return True


def upload_directory_to_s3(client, path, bucket_name, s3_prefix="input"):
    if not os.path.exists(path):
        return False
    for root, dirs, files in os.walk(path):
        for filename in files:
            local_path = os.path.join(root, filename)
            # generate s3 dir path
            # Skip the first /
            # HACK
            s3_path = local_path.split(path)[1][1:]
            # Upload
            if s3_prefix:
                s3_path = os.path.join(s3_prefix, s3_path)
            try:
                client.upload_file(local_path, bucket_name, s3_path)
            except Exception:
                return False
    return True


def bucket_exists(s3_client, bucket_name):
    try:
        s3_client.head_bucket(Bucket=bucket_name)
        return True
    except botocore.exceptions.ClientError:
        pass
    return False


def create_bucket(s3_client, bucket_name, **kwargs):
    bucket = s3_client.create_bucket(Bucket=bucket_name, **kwargs)
    if not bucket:
        return False
    return bucket


def delete_bucket(s3_client, bucket_name, **kwargs):
    return s3_client.delete_bucket(Bucket=bucket_name, **kwargs)


def delete_objects(s3_resource, bucket_name):
    bucket = s3_resource.Bucket(bucket_name)
    objects_keys = [{"Key": obj.key} for obj in bucket.objects.all()]
    return s3_resource.meta.client.delete_objects(
        Bucket=bucket_name, Delete={"Objects": objects_keys}
    )


# Accept parameters to
def expand_s3_bucket(s3_resource, bucket_name, target_dir=None, prefix="input"):
    """ Expands the content of the specified bucket into the target_dir """
    bucket = s3_resource.Bucket(bucket_name)
    for obj in bucket.objects.filter(Prefix=prefix):
        obj_path = copy.deepcopy(obj.key)
        if prefix:
            obj_path = os.path.relpath(obj_path, prefix)

        if target_dir:
            full_path = os.path.join(target_dir, obj_path)
        else:
            full_path = obj_path
        dir_path = os.path.dirname(full_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        bucket.download_file(obj.key, full_path)
    return True


def stage_s3_resource(**kwargs):
    return boto3.resource("s3", **kwargs)



import unittest
import os
from jobio.util import validate_dict_types, validate_dict_values, create_dir
from jobio.storage.staging import required_staging_fields, required_staging_values

from jobio.storage.s3 import (
    required_s3_fields,
    required_s3_values,
    required_bucket_fields,
    required_bucket_values,
    upload_directory_to_s3,
    expand_s3_bucket,
    stage_s3_resource,
    create_bucket,
    delete_bucket,
    delete_objects,
)


here = os.path.abspath(__file__)
current_dir = os.path.dirname(here)


class TestStorage(unittest.TestCase):
    def setUp(self):
        self.staging_options = dict(
            enable=True,
            secrets_dir="",
            input_path="",
            output_path="results",
            endpoint="https://ku.compat.objectstorage.eu-frankfurt-1"
            ".oraclecloud.com",
        )

        self.s3_options = dict(region_name="eu-frankfurt-1")

        self.bucket_options = dict(
            name="bucket_test_name", input_prefix="input", output_prefix="output",
        )

    def tearDown(self):
        self.staging_options = None
        self.s3_options = None

    def test_staging_options(self):
        self.assertTrue(
            validate_dict_types(
                self.staging_options, required_fields=required_staging_fields,
            )
        )

        self.assertTrue(
            validate_dict_values(
                self.staging_options, required_values=required_staging_values
            )
        )

    def test_s3_options(self):
        self.assertTrue(
            validate_dict_types(self.s3_options, required_fields=required_s3_fields)
        )

        self.assertTrue(
            validate_dict_values(self.s3_options, required_values=required_s3_values)
        )

    def test_bucket_options(self):
        self.assertTrue(
            validate_dict_types(
                self.bucket_options, required_fields=required_bucket_fields
            )
        )

        self.assertTrue(
            validate_dict_values(
                self.bucket_options, required_values=required_bucket_values
            )
        )

    def test_s3_resource(self):
        s3_resource = stage_s3_resource(
            endpoint_url=self.staging_options["endpoint"], **self.s3_options
        )
        self.assertIsNotNone(s3_resource)

    def test_create_delete_bucket(self):
        s3_resource = stage_s3_resource(
            endpoint_url=self.staging_options["endpoint"], **self.s3_options
        )
        self.assertIsNotNone(s3_resource)

        bucket = create_bucket(s3_resource.meta.client, self.bucket_options["name"])

        self.assertIsInstance(bucket, dict)
        deleted = delete_bucket(s3_resource.meta.client, self.bucket_options["name"])

        self.assertIsInstance(deleted, dict)
        self.assertEqual(204, deleted["ResponseMetadata"]["HTTPStatusCode"])

    def test_upload_expand_bucket(self):
        # Expand test directory
        s3_resource = stage_s3_resource(
            endpoint_url=self.staging_options["endpoint"], **self.s3_options
        )

        bucket_options = dict(
            name="jobio", input_prefix="input", output_prefix="output",
        )
        local_directory = os.path.join(current_dir, "res", "job_input")

        bucket = create_bucket(s3_resource.meta.client, bucket_options["name"])
        self.assertIsInstance(bucket, dict)

        uploaded = upload_directory_to_s3(
            s3_resource.meta.client, local_directory, bucket_options["name"]
        )
        self.assertTrue(uploaded)

        expand_dir = os.path.join(current_dir, "res", "expand_dir")
        if not os.path.exists(expand_dir):
            self.assertTrue(create_dir(expand_dir))

        expaned = expand_s3_bucket(s3_resource, bucket_options["name"], expand_dir)
        self.assertTrue(expaned)

        deleted = delete_objects(s3_resource, bucket_options["name"])

        # Delete all content
        deleted = delete_bucket(s3_resource.meta.client, bucket_options["name"])
        self.assertIsInstance(deleted, dict)
        self.assertEqual(204, deleted["ResponseMetadata"]["HTTPStatusCode"])



import boto3
import os
import subprocess
import time
from jobio.cli.args import extract_arguments
from jobio.defaults import EXECUTE, JOB, S3, BUCKET, STORAGE, JOB_DEFAULT_NAME
from jobio.storage.staging import required_staging_fields, required_staging_values
from jobio.storage.s3 import (
    bucket_exists,
    create_bucket,
    expand_s3_bucket,
    valid_s3_resource_fields,
    required_bucket_fields,
    required_bucket_values,
    required_s3_fields,
    required_s3_values,
    upload_directory_to_s3,
)
from jobio.util import (
    validate_dict_types,
    validate_dict_values,
    remove_dir,
    save_results,
    load_kubernetes_secrets,
)


required_execute_fields = {
    "command": str,
    "args": list,
    "capture": bool,
    "output_path": str,
}

required_execute_values = {
    "command": True,
    "args": False,
    "capture": False,
    "output_path": False,
}


def process(execute_kwargs=None):
    if not execute_kwargs:
        execute_kwargs = {}

    command = [execute_kwargs["command"]]
    if "args" in execute_kwargs:
        command.extend(execute_kwargs["args"])

    # Subprocess
    result = subprocess.run(command, capture_output=execute_kwargs["capture"])
    output_results = {}
    if hasattr(result, "args"):
        output_results.update({"command": " ".join((getattr(result, "args")))})
    if hasattr(result, "returncode"):
        output_results.update({"returncode": str(getattr(result, "returncode"))})
    if hasattr(result, "stderr"):
        output_results.update({"error": str(getattr(result, "stderr"))})
    if hasattr(result, "stdout"):
        output_results.update({"output": str(getattr(result, "stdout"))})
    return output_results


def submit(args):
    job_dict = vars(extract_arguments(args, [JOB]))
    if "name" not in job_dict or not job_dict["name"]:
        since_epoch = int(time.time())
        job_dict["name"] = "{}-{}".format(JOB_DEFAULT_NAME, since_epoch)

    execute_dict = vars(extract_arguments(args, [EXECUTE]))
    valid_job_types = validate_dict_types(
        execute_dict, required_fields=required_execute_fields, verbose=job_dict["debug"]
    )
    valid_job_values = validate_dict_values(
        execute_dict, required_values=required_execute_values, verbose=job_dict["debug"]
    )

    if not valid_job_types:
        raise TypeError(
            "Incorrect required executable arguments were provided: {}".format(
                valid_job_types
            )
        )
    if not valid_job_values:
        raise ValueError(
            "Missing required executable values: {}".format(valid_job_values)
        )

    staging_storage_dict = vars(extract_arguments(args, [STORAGE]))
    s3_dict = vars(extract_arguments(args, [S3]))
    bucket_dict = vars(extract_arguments(args, [BUCKET]))

    # Dynamically get secret credentials
    if staging_storage_dict["enable"]:

        # Validate bucket arguments
        validate_dict_types(
            bucket_dict,
            required_fields=required_bucket_fields,
            verbose=job_dict["debug"],
            throw=True,
        )

        validate_dict_values(
            bucket_dict,
            required_values=required_bucket_values,
            verbose=job_dict["debug"],
            throw=True,
        )

        validate_dict_types(
            staging_storage_dict,
            required_fields=required_staging_fields,
            verbose=job_dict["debug"],
            throw=True,
        )
        validate_dict_values(
            staging_storage_dict,
            required_values=required_staging_values,
            verbose=job_dict["debug"],
            throw=True,
        )

        # Validate staging arguments
        load_secrets = ["aws_access_key_id", "aws_secret_access_key"]
        loaded_secrests = load_kubernetes_secrets(
            staging_storage_dict["secrets_dir"], load_secrets
        )
        for k, v in loaded_secrests.items():
            s3_dict.update({k: v})

        validate_dict_types(
            s3_dict,
            required_fields=required_s3_fields,
            verbose=job_dict["debug"],
            throw=True,
        )
        validate_dict_values(
            s3_dict,
            required_values=required_s3_values,
            verbose=job_dict["debug"],
            throw=True,
        )

        # Only used valid_s3_fields
        resources_fields = {
            k: v for k, v in s3_dict.items() if k in valid_s3_resource_fields
        }

        s3_resource = boto3.resource("s3", **resources_fields)
        # Load aws credentials
        expanded = expand_s3_bucket(
            s3_resource,
            bucket_dict["name"],
            target_dir=staging_storage_dict["input_path"],
        )
        if not expanded:
            raise RuntimeError(
                "Failed to expand the target bucket: {}".format(bucket_dict["name"])
            )

    result = process(execute_kwargs=execute_dict)
    saved = False
    final_result_path = ""

    # Put results into the put path
    if "output_path" in execute_dict:
        root, ext = os.path.splitext(execute_dict["output_path"])
        if ext:
            final_result_path = execute_dict["output_path"]
        else:
            # A directory path
            # Generate a filename to put inside the directory
            result_output_file = "{}.txt".format(job_dict["name"])
            final_result_path = os.path.join(
                execute_dict["output_path"], result_output_file
            )
        if final_result_path:
            saved = save_results(final_result_path, result)
            if not saved:
                raise RuntimeError(
                    "Failed to save the results of job: {}".format(job_dict["name"])
                )

    if staging_storage_dict["enable"]:
        if not bucket_exists(s3_resource.meta.client, bucket_dict["name"]):
            # TODO, load region from AWS config
            created = create_bucket(
                s3_resource.meta.client,
                bucket_dict["name"],
                CreateBucketConfiguration={"LocationConstraint": s3_dict["region"]},
            )
            if not created:
                raise RuntimeError(
                    "Failed to create results bucket: {}".format(bucket_dict["name"])
                )

        uploaded = upload_directory_to_s3(
            s3_resource.meta.client,
            staging_storage_dict["output_path"],
            bucket_dict["name"],
        )
        if not uploaded:
            print("Failed to upload results")

        # Cleanout local results
        if os.path.exists(os.path.dirname(final_result_path)):
            if not remove_dir(os.path.dirname(final_result_path)):
                print("Failed to remove results after upload")
        # TODO, cleanout inputs
        if os.path.exists(staging_storage_dict["input_path"]):
            if not remove_dir(staging_storage_dict["input_path"]):
                print("Failed to remove input after upload")

    print("{}".format(result))
    # TODO, print output to stdout



import argparse
from argparse import Namespace

from jobio.defaults import EXECUTE, JOB, S3, STORAGE, BUCKET


def strip_argument_prefix(arguments, prefix=""):
    return {k.replace(prefix, ""): v for k, v in arguments.items()}


def _get_arguments(arguments, startswith=""):
    return {k: v for k, v in arguments.items() if k.startswith(startswith)}


def add_job_meta_group(parser):
    meta_group = parser.add_argument_group(title="Job metadata")
    meta_group.add_argument("--job-name", default=False)
    meta_group.add_argument("--job-debug", action="store_true", default=False)


def add_execute_group(parser):
    execute_group = parser.add_argument_group(title="Execute arguments")
    execute_group.add_argument("execute_command", default="")
    execute_group.add_argument("--execute-args", nargs="*", default=[])
    execute_group.add_argument("--execute-capture", action="store_true", default=True)
    execute_group.add_argument("--execute-output-path", default="")


def add_bucket_group(parser):
    bucket_group = parser.add_argument_group(title="Bucket arguments")
    bucket_group.add_argument("--bucket-name", default="")
    bucket_group.add_argument("--bucket-input-prefix", default="input")
    bucket_group.add_argument("--bucket-output-prefix", default="output")


def add_s3_group(parser):
    s3_group = parser.add_argument_group(title="S3 arguments")
    s3_group.add_argument("--s3-profile-name", default="default")
    s3_group.add_argument("--s3-region-name", default="eu-frankfurt-1")


def add_storage_group(parser):
    storage_group = parser.add_argument_group(title="Staging Storage arguments")
    storage_providers = storage_group.add_mutually_exclusive_group()
    storage_providers.add_argument("--storage-s3", action="store_true")
    storage_group.add_argument("--storage-enable", action="store_true", default=False)
    storage_group.add_argument("--storage-endpoint", default="")
    storage_group.add_argument("--storage-secrets-dir", default="")
    storage_group.add_argument("--storage-input-path", default="/tmp/input")
    storage_group.add_argument("--storage-output-path", default="/tmp/output")


argument_groups = {
    EXECUTE: add_execute_group,
    JOB: add_job_meta_group,
    STORAGE: add_storage_group,
    S3: add_s3_group,
    BUCKET: add_bucket_group,
}


def get_arguments(argument_types, strip_group_prefix=True):
    parser = argparse.ArgumentParser()

    for argument_group in argument_types:
        if argument_group in argument_groups:
            argument_groups[argument_group](parser)

    args, unknown = parser.parse_known_args()
    if strip_group_prefix:
        stripped_args = {}
        for argument_group in argument_types:
            group_args = _get_arguments(vars(args), argument_group.lower())
            group_args = strip_argument_prefix(group_args, argument_group.lower() + "_")
            stripped_args.update(group_args)
        return Namespace(**stripped_args)
    return args


def extract_arguments(arguments, argument_types, strip_group_prefix=True):
    if strip_group_prefix:
        stripped_args = {}
        for argument_group in argument_types:
            group_args = _get_arguments(vars(arguments), argument_group.lower())
            group_args = strip_argument_prefix(group_args, argument_group.lower() + "_")
            stripped_args.update(group_args)
        return Namespace(**stripped_args)
    return {}



required_staging_fields = {
    "enable": bool,
    "endpoint": str,
    "secrets_dir": str,
    "input_path": str,
    "output_path": str,
}

required_staging_values = {
    "enable": True,
    "endpoint": True,
    "secrets_dir": False,
    "input_path": False,
    "output_path": True,
}



import argparse
from jobio.defaults import PACKAGE_NAME
from jobio.cli.helpers import add_job_cli


def run():
    parser = argparse.ArgumentParser(prog=PACKAGE_NAME)
    add_job_cli(parser)

    args = parser.parse_args()
    # Execute default funciton
    if hasattr(args, "func"):
        result = args.func(args)
        if result:
            print(result)
    return None


if __name__ == "__main__":
    arguments = run()



import unittest
from jobio.job import process


class TestJOB(unittest.TestCase):
    def test_process(self):
        execute_kwargs = dict(command="/bin/echo", args=["Hello World"], capture=True)
        results = process(execute_kwargs=execute_kwargs)
        expected_results = dict(
            command="/bin/echo Hello World",
            returncode="0",
            error="b''",
            output="b'Hello World\\n'",
        )

        self.assertDictEqual(results, expected_results)



#!/usr/bin/python
# coding: utf-8
import os
from setuptools import setup, find_packages

cur_dir = os.path.abspath(os.path.dirname(__file__))


def read(path):
    with open(path, "r") as _file:
        return _file.read()


def read_req(name):
    path = os.path.join(cur_dir, name)
    return [req.strip() for req in read(path).splitlines() if req.strip()]


version_ns = {}
with open(os.path.join(cur_dir, "version.py")) as f:
    exec(f.read(), {}, version_ns)


long_description = open("README.rst").read()
setup(
    name="jobio",
    version=version_ns["__version__"],
    description="A bootstrap package to handle input and "
    "output data for batch processes",
    long_description=long_description,
    author="Rasmus Munk",
    author_email="munk1@live.dk",
    packages=find_packages(),
    url="https://github.com/rasmunk/jobio",
    license="MIT",
    keywords=["Job", "IO", "S3", "Batch"],
    install_requires=read_req("requirements.txt"),
    extras_require={
        "test": read_req("requirements-dev.txt"),
        "dev": read_req("requirements-dev.txt"),
    },
    entry_points={"console_scripts": ["jobio = jobio.cli.cli:run"]},
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import os
import json
import shutil


def create_dir(path):
    try:
        os.makedirs(path)
        return True
    except IOError as err:
        print("Failed to create: {}, err: {}".format(path, err))
    return False


def remove_dir(path):
    try:
        shutil.rmtree(path)
        return True
    except OSError as err:
        print("Failed to remove: {}, err: {}".format(path, err))
    return False


def present_in(var, collection, verbose=False):
    if var not in collection:
        if verbose:
            print("variable: {} not present in: {}".format(var, collection))
        return False
    return True


def validate_dict_types(input_dict, required_fields=None, verbose=False, throw=False):
    if not required_fields:
        required_fields = {}
    for var, _type in required_fields.items():
        if not present_in(var, input_dict, verbose=verbose):
            if throw:
                raise KeyError(
                    "Missing required: {} in {}".format(var, required_fields)
                )
            return False
        if not isinstance(input_dict[var], _type):
            if verbose:
                print(
                    "variable: {} value: {} is of incorrect type: {}".format(
                        var, input_dict[var], type(input_dict[var])
                    )
                )
            if throw:
                raise TypeError("{}: should be {}".format(var, _type))
            return False
    return True


def validate_dict_values(input_dict, required_values=None, verbose=False, throw=False):
    if not required_values:
        required_values = {}

    for var, required_value in required_values.items():
        if not present_in(var, input_dict, verbose=verbose):
            if throw:
                raise KeyError("Missing required: {} in {} ".format(var, input_dict))
            return False
        if required_value and not input_dict[var]:
            if verbose:
                print(
                    "The required variable: {} was not set in: {}".format(
                        var, input_dict
                    )
                )
            if throw:
                raise ValueError(
                    "{} doesn't have a value: {}".format(var, required_value)
                )
            return False
    return True


def save_results(path, results):
    save_dir = os.path.dirname(path)
    if save_dir and not os.path.exists(save_dir):
        if not create_dir(save_dir):
            return False
    try:
        with open(path, "w") as fh:
            try:
                json.dump(results, fh)
            except TypeError as j_err:
                print("Failed to serialize to json: {}".format(j_err))
                return False
        return True
    except IOError as err:
        print("Failed to save results: {}".format(err))
    return False


def load_kubernetes_secrets(directory, secret_keys, strip_file_newline=True):
    loaded_secrets = {}
    for secret_key in secret_keys:
        value_path = os.path.join(directory, secret_key)
        if os.path.exists(value_path):
            if os.path.islink(value_path):
                value_path = os.path.realpath(value_path)
            if os.path.isfile(value_path) and not os.path.islink(value_path):
                content = None
                try:
                    with open(value_path, "rb") as fh:
                        content = fh.read()
                except IOError as err:
                    print("Failed to read file: {}".format(err))
                decoded = content.decode("utf-8")
                if strip_file_newline:
                    decoded = decoded.replace("\n", "")
                loaded_secrets[secret_key] = decoded
    return loaded_secrets
